/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <gnomevfs-mount.h>

static GnomeVFSURI *base_uri = NULL;
static GHashTable *opened_files_hash = NULL;
static GHashTable *file_info_files_hash = NULL;
static gboolean __debug = FALSE;

static void gnome_vfs_mount_init() {
	gnome_vfs_init();
	
	opened_files_hash = g_hash_table_new_full(
		g_str_hash,
		g_str_equal,
		g_free,
		(GDestroyNotify)gnome_vfs_close);

	file_info_files_hash = g_hash_table_new_full(
		g_str_hash,
		g_str_equal,
		g_free,
		(GDestroyNotify)gnome_vfs_file_info_unref);		
}

static void gnome_vfs_mount_shutdown() {
	gnome_vfs_uri_unref(base_uri);
	g_hash_table_destroy(file_info_files_hash);
	g_hash_table_destroy(opened_files_hash);
	gnome_vfs_shutdown();
}

static void gnome_vfs_mount_critical(GnomeVFSResult code, const gchar *path) {
	const gchar *error_desc;
	if(!__debug)
		return;
	error_desc = gnome_vfs_result_to_string(code);
	g_critical("Error %d when accessing %s: %s\n", code, path, error_desc);
}

static void gnome_vfs_mount_message(gchar const *format, ...) {
	gchar *message;
	va_list args;
	if(!__debug)
		return;
	va_start(args, format);
	g_vasprintf(&message, format, args);
	g_message(message);
	g_free(message);
	va_end(args);
}

static GnomeVFSResult gnome_vfs_mount_get_file_info(
			const gchar *path,
			GnomeVFSFileInfo *info,
			GnomeVFSFileInfoOptions options) {
	GnomeVFSResult result;
	GnomeVFSFileInfo *cache_info;
	
	cache_info = (GnomeVFSFileInfo *)g_hash_table_lookup(file_info_files_hash, path);

	if(cache_info == NULL) {
		GnomeVFSURI *uri;
		gnome_vfs_mount_message("%s not found in cache\n", path);
		uri = gnome_vfs_uri_append_path(base_uri, path);
		cache_info = gnome_vfs_file_info_new();
		result = gnome_vfs_get_file_info_uri(uri, cache_info, options);
		if(result == GNOME_VFS_OK)
			g_hash_table_replace(file_info_files_hash, g_strdup(path), cache_info);
		gnome_vfs_uri_unref(uri);
	}
	else {
		result = GNOME_VFS_OK;
		gnome_vfs_mount_message("%s found in cache\n", path);
	}

	gnome_vfs_file_info_copy(info, cache_info);
	return result;
}

static gboolean gnome_vfs_mount_file_info_forget(const gchar *path) {
	gboolean was_removed;

	was_removed = g_hash_table_remove(file_info_files_hash, path);
	if(was_removed)
		gnome_vfs_mount_message("Forgetting %s\n", path);
	else
		gnome_vfs_mount_message("Could not forget %s\n", path);

	return was_removed;
}

static int gnome_vfs_mount_getattr(const char *path, struct stat *stbuf) {
	int res = 0;
	int code = 0;
	GnomeVFSFileInfo *info = NULL;
	
	gnome_vfs_mount_message("gnome_vfs_mount_getattr %s\n", path);

	memset(stbuf, 0, sizeof(struct stat));

	info = gnome_vfs_file_info_new();
	code = gnome_vfs_mount_get_file_info(path, info, GNOME_VFS_FILE_INFO_DEFAULT);

	if(code == GNOME_VFS_ERROR_NOT_FOUND)
		res = -ENOENT;
	else
	if(code == GNOME_VFS_ERROR_PROTOCOL_ERROR)
		res = -ENOENT;
	else
	if(code == GNOME_VFS_OK) {
		stbuf->st_uid = getuid();
		stbuf->st_gid = getgid();

		if(info->valid_fields & GNOME_VFS_FILE_INFO_FIELDS_TYPE)
			switch(info->type) {
				case GNOME_VFS_FILE_TYPE_REGULAR:
					stbuf->st_mode |= S_IFREG;
					break;
				case GNOME_VFS_FILE_TYPE_DIRECTORY:
					stbuf->st_mode |= S_IFDIR;
					break;
				case GNOME_VFS_FILE_TYPE_FIFO:
					stbuf->st_mode |= S_IFIFO;
					break;
				case GNOME_VFS_FILE_TYPE_SOCKET:
					stbuf->st_mode |= S_IFSOCK;
					break;
				case GNOME_VFS_FILE_TYPE_CHARACTER_DEVICE:
					stbuf->st_mode |= S_IFCHR;
					break;
				case GNOME_VFS_FILE_TYPE_BLOCK_DEVICE:
					stbuf->st_mode |= S_IFBLK;
					break;
				case GNOME_VFS_FILE_TYPE_SYMBOLIC_LINK:
					stbuf->st_mode |= S_IFLNK;
					break;
				case GNOME_VFS_FILE_TYPE_UNKNOWN:
					break;
			}

		if(info->valid_fields & GNOME_VFS_FILE_INFO_FIELDS_PERMISSIONS)
			stbuf->st_mode |= info->permissions;
		else {
			if(stbuf->st_mode & S_IFREG)
				stbuf->st_mode |= 0440;
			else if(stbuf->st_mode & S_IFDIR)
				stbuf->st_mode |= 0550;
		}

		if(info->valid_fields & GNOME_VFS_FILE_INFO_FIELDS_FLAGS)
			if (info->flags & GNOME_VFS_FILE_FLAGS_LOCAL) {

				if(info->valid_fields & GNOME_VFS_FILE_INFO_FIELDS_DEVICE)
					stbuf->st_dev = info->device;

				if(info->valid_fields & GNOME_VFS_FILE_INFO_FIELDS_INODE)
					stbuf->st_ino = info->inode;
			}

		if(info->valid_fields & GNOME_VFS_FILE_INFO_FIELDS_SIZE)
			stbuf->st_size = info->size;

		if(info->valid_fields & GNOME_VFS_FILE_INFO_FIELDS_LINK_COUNT)
			stbuf->st_nlink = info->link_count;

		if(info->valid_fields & GNOME_VFS_FILE_INFO_FIELDS_BLOCK_COUNT)
			stbuf->st_blocks = info->block_count;

		if(info->valid_fields & GNOME_VFS_FILE_INFO_FIELDS_IO_BLOCK_SIZE)
			stbuf->st_blksize = info->io_block_size;

		if(info->valid_fields & GNOME_VFS_FILE_INFO_FIELDS_ATIME)
			stbuf->st_atime = info->atime;

		if(info->valid_fields & GNOME_VFS_FILE_INFO_FIELDS_MTIME)
			stbuf->st_mtime = info->mtime;

		if(info->valid_fields & GNOME_VFS_FILE_INFO_FIELDS_CTIME)
			stbuf->st_ctime = info->ctime;
	}
	else {
		res = -ENOENT;
		gnome_vfs_mount_critical(code, path);
	}

	gnome_vfs_mount_message(
			"n_link = %u, size = %lu, blocks = %lu, blksize = %lu\n",
			stbuf->st_nlink,
			stbuf->st_size,
			stbuf->st_blocks,
			stbuf->st_blksize
		);

	gnome_vfs_file_info_unref(info);	
	return res;
}

static int gnome_vfs_mount_readlink(const char *path, char *buf, size_t size) {
	int res = 0;
	int code = 0;
	GnomeVFSFileInfo *info = NULL;

	info = gnome_vfs_file_info_new();
	code = gnome_vfs_mount_get_file_info(path, info, GNOME_VFS_FILE_INFO_DEFAULT);
	if(code != GNOME_VFS_OK)
		gnome_vfs_mount_critical(code, path);

	if(info->valid_fields & GNOME_VFS_FILE_INFO_FIELDS_SYMLINK_NAME) {
		gnome_vfs_mount_message(
			"gnome_vfs_mount_readlink %s -> %s\n",
			path, info->symlink_name);
		strncpy(buf, info->symlink_name, size);
	}
	else
		res = -EINVAL;

	gnome_vfs_file_info_unref(info);
	return res;
}

static int gnome_vfs_mount_getdir(
			const char *path,
			fuse_dirh_t h,
			fuse_dirfil_t filler) {
	int res = 0;
	int code = 0;
	GnomeVFSURI *uri = NULL;
	GnomeVFSDirectoryHandle *dir = NULL;
	GnomeVFSFileInfo *info = NULL;

	gnome_vfs_mount_message("gnome_vfs_mount_getdir %s\n", path);

	uri = gnome_vfs_uri_append_path(base_uri, path);
	info = gnome_vfs_file_info_new();
	code = gnome_vfs_directory_open_from_uri(
		&dir, uri,
		GNOME_VFS_FILE_INFO_DEFAULT);

	if(code != GNOME_VFS_OK) {
		gnome_vfs_mount_critical(code, path);
		res = -ENOENT;
	}
	else {
		for(;;) {
			code = gnome_vfs_directory_read_next(dir, info);
			if(code != GNOME_VFS_OK) {
				if(code == GNOME_VFS_ERROR_EOF)
					break;
				else
					gnome_vfs_mount_critical(code, path);
			}
			filler(h, info->name, 0);
		}
		gnome_vfs_directory_close(dir);
	}

	gnome_vfs_file_info_unref(info);
	gnome_vfs_uri_unref(uri);
	return res;
}

static int gnome_vfs_mount_open(const char *path, int flags) {
/*
O_RDONLY   =    0
O_WRONLY   =    1
O_RDWR     =    2
O_CREAT    =   64
O_EXCL     =  128
O_NOCTTY   =  256
O_TRUNC    =  512
O_APPEND   = 1024
O_NONBLOCK = 2048
O_SYNC     = 4096
*/
	int res = 0;
	int code = 0;
	GnomeVFSURI *uri = NULL;
	GnomeVFSHandle *handle = NULL;
	GnomeVFSOpenMode mode = GNOME_VFS_OPEN_RANDOM;

	gnome_vfs_mount_message("gnome_vfs_mount_open %s\n", path);

	switch(flags & (O_RDONLY|O_WRONLY|O_RDWR) ) {
		case O_RDONLY:
			mode |= GNOME_VFS_OPEN_READ;
			break;
		case O_WRONLY:
			mode |= GNOME_VFS_OPEN_WRITE;
			break;
		case O_RDWR:
			mode |= (GNOME_VFS_OPEN_WRITE | GNOME_VFS_OPEN_READ);
	}

	uri = gnome_vfs_uri_append_path(base_uri, path);

	code = gnome_vfs_open_uri(&handle, uri, mode);
	if(code == GNOME_VFS_ERROR_NOT_FOUND)
		res = -ENOENT;
	else
	if(code == GNOME_VFS_ERROR_NOT_SUPPORTED)
		res = -EINVAL;
	else
	if(code != GNOME_VFS_OK) {
		res = -EINVAL;
		gnome_vfs_mount_critical(code, path);
	}
	else
		g_hash_table_replace(opened_files_hash, g_strdup(path), handle);
	
	gnome_vfs_uri_unref(uri);
	return res;
}

static int gnome_vfs_mount_read(const char *path, char *buf, size_t size, off_t offset) {
	int code = 0;
	int res = 0;
	GnomeVFSURI *uri = NULL;
	GnomeVFSHandle *handle = NULL;
	GnomeVFSFileSize bytes_read = 0;
	GnomeVFSFileSize total_bytes_read = 0;
	GnomeVFSFileSize told;

	gnome_vfs_mount_message("gnome_vfs_mount_read %s %d %lu\n", path, size, offset);

	uri = gnome_vfs_uri_append_path(base_uri, path);
	handle = (GnomeVFSHandle *)g_hash_table_lookup(opened_files_hash, path);
	
	if(handle == NULL)
		res = -ENOENT;
	else {
		gnome_vfs_tell(handle, &told);
		if(code != GNOME_VFS_OK) {
			res = -EIO;
			gnome_vfs_mount_critical(code, path);
		}
		if(told != offset)
			code = gnome_vfs_seek(handle, GNOME_VFS_SEEK_START, offset);
		if(code != GNOME_VFS_OK) {
			res = -EIO;
			gnome_vfs_mount_critical(code, path);
		}
		do {
			code = gnome_vfs_read(
						handle, buf + total_bytes_read,
						size, &bytes_read);
						
			size -= bytes_read;
			total_bytes_read += bytes_read;
			if(code != GNOME_VFS_OK && code != GNOME_VFS_ERROR_EOF) {
				res = -EIO;
				gnome_vfs_mount_critical(code, path);
				break;
			}
		} while(size > 0 && bytes_read != 0);
	}

	gnome_vfs_mount_message("%llu bytes read\n", total_bytes_read);
	gnome_vfs_uri_unref(uri);
	return total_bytes_read;
}

static int gnome_vfs_mount_write(
		const char *path, const char *buf,
		size_t size, off_t offset) {
	int code = 0;
	int res = 0;
	GnomeVFSURI *uri = NULL;
	GnomeVFSHandle *handle = NULL;
	GnomeVFSFileSize bytes_wrote = 0;
	GnomeVFSFileSize total_bytes_wrote = 0;
	GnomeVFSFileSize told;
	
	gnome_vfs_mount_message(
		"gnome_vfs_mount_write %s %d %lu\n",
		path, size, offset);

	uri = gnome_vfs_uri_append_path(base_uri, path);
	gnome_vfs_mount_file_info_forget(path);
	handle = (GnomeVFSHandle *)g_hash_table_lookup(opened_files_hash, path);
	
	if(handle == NULL)
		res = -ENOENT;
	else {
		if(offset != 0) {
			code = gnome_vfs_tell(handle, &told);
			if(code != GNOME_VFS_OK) {
				res = -EIO;
				gnome_vfs_mount_critical(code, path);
			}
			if(told != offset)
				code = gnome_vfs_seek(handle, GNOME_VFS_SEEK_START, offset);
			if(code != GNOME_VFS_OK) {
				res = -EIO;
				gnome_vfs_mount_critical(code, path);
			}
		}
		
		do {
			code = gnome_vfs_write(
						handle,buf + total_bytes_wrote,
						size, &bytes_wrote);
			size -= bytes_wrote;
			total_bytes_wrote += bytes_wrote;
			if(code != GNOME_VFS_OK && code != GNOME_VFS_ERROR_EOF) {
				res = -EIO;
				gnome_vfs_mount_critical(code, path);
				break;
			}
		} while(size > 0 && bytes_wrote != 0);
	}
	
	gnome_vfs_mount_message("%llu bytes wrote\n", total_bytes_wrote);
	gnome_vfs_uri_unref(uri);
	return total_bytes_wrote;
}

static int gnome_vfs_mount_truncate(const char *path, off_t size) {
	int code = 0;
	int res = 0;
	GnomeVFSURI *uri = NULL;

	gnome_vfs_mount_message("gnome_vfs_mount_truncate %s %lu\n", path, size);
	
	uri = gnome_vfs_uri_append_path(base_uri, path);
	gnome_vfs_mount_file_info_forget(path);
	
	code = gnome_vfs_truncate_uri(uri, size);
	if(code == GNOME_VFS_ERROR_NOT_FOUND)
		res = -ENOENT;
	else
	if(code == GNOME_VFS_ERROR_NOT_SUPPORTED)
		res = -EACCES;
	else
	if(code != GNOME_VFS_OK) {
		res = -EACCES;
		gnome_vfs_mount_critical(code, path);
	}
	
	gnome_vfs_uri_unref(uri);
	return res;
}

static int gnome_vfs_mount_mkdir(const char *path, mode_t mode) {
	int code = 0;
	int res = 0;
	GnomeVFSURI *uri = NULL;

	gnome_vfs_mount_message("gnome_vfs_mount_mkdir %s %o\n", path, mode);

	uri = gnome_vfs_uri_append_path(base_uri, path);
	code = gnome_vfs_make_directory_for_uri(uri, mode);

	if(code == GNOME_VFS_ERROR_NOT_SUPPORTED)
		res = -EACCES;
	else	
	if(code != GNOME_VFS_OK) {
		res = -EACCES;
		gnome_vfs_mount_critical(code, path);
	}

	gnome_vfs_uri_unref(uri);
	return res;
}

static int gnome_vfs_mount_rmdir(const char *path) {
	int code = 0;
	int res = 0;
	GnomeVFSURI *uri = NULL;
	
	gnome_vfs_mount_message("gnome_vfs_mount_rmdir %s\n", path);
	
	uri = gnome_vfs_uri_append_path(base_uri, path);
	gnome_vfs_mount_file_info_forget(path);
	code = gnome_vfs_remove_directory_from_uri(uri);
	
	if(code == GNOME_VFS_ERROR_NOT_SUPPORTED)
		res = -EACCES;
	else	
	if(code != GNOME_VFS_OK) {
		res = -EACCES;
		gnome_vfs_mount_critical(code, path);
	}
	
	gnome_vfs_uri_unref(uri);
	return res;
}

static int gnome_vfs_mount_unlink(const char *path) {
	int code = 0;
	int res = 0;
	GnomeVFSURI *uri = NULL;

	gnome_vfs_mount_message("gnome_vfs_mount_unlink %s\n", path);
	
	uri = gnome_vfs_uri_append_path(base_uri, path);
	gnome_vfs_mount_file_info_forget(path);
	code = gnome_vfs_unlink_from_uri(uri);
	
	if(code == GNOME_VFS_ERROR_NOT_SUPPORTED)
		res = -EACCES;
	else	
	if(code != GNOME_VFS_OK) {
		res = -EACCES;
		gnome_vfs_mount_critical(code, path);
	}

	gnome_vfs_uri_unref(uri);
	return res;
}

static int gnome_vfs_mount_utime(const char *path, struct utimbuf *buf) {
	int code = 0;
	int res = 0;
	GnomeVFSURI *uri = NULL;
	GnomeVFSFileInfo *info = NULL;

	gnome_vfs_mount_message("gnome_vfs_mount_utime %s\n", path);
	
	uri = gnome_vfs_uri_append_path(base_uri, path);
	gnome_vfs_mount_file_info_forget(path);
	info = gnome_vfs_file_info_new();
	
	if(buf == NULL) {
		time_t now;
		now = time(NULL);
		info->atime = now;
		info->mtime = now;
	}
	else {
		info->atime = buf->actime;
		info->mtime = buf->modtime;
	}

	code = gnome_vfs_set_file_info_uri(uri, info, GNOME_VFS_SET_FILE_INFO_TIME);
	if(code == GNOME_VFS_ERROR_NOT_SUPPORTED)
		res = -EACCES;
	else	
	if(code != GNOME_VFS_OK) {
		res = -EACCES;
		gnome_vfs_mount_critical(code, path);
	}
	
	gnome_vfs_uri_unref(uri);
	gnome_vfs_file_info_unref(info);
	return res;
}

static int gnome_vfs_mount_mknod(const char *path, mode_t mode, dev_t rdev) {
	int code = 0;
	int res = 0;
	GnomeVFSURI *uri = NULL;
	GnomeVFSHandle *handle = NULL;

	gnome_vfs_mount_message("gnome_vfs_mount_mknod %s %o\n", path, mode);

	uri = gnome_vfs_uri_append_path(base_uri, path);
	code = gnome_vfs_create_uri(&handle, uri, GNOME_VFS_OPEN_WRITE, TRUE, mode);
	
	if(code == GNOME_VFS_ERROR_NOT_SUPPORTED)
		res = -EACCES;
	else
	if(code == GNOME_VFS_ERROR_ACCESS_DENIED)
		res = -EACCES;
	else
	if(code != GNOME_VFS_OK) {
		res = -EACCES;
		gnome_vfs_mount_critical(code, path);
	}

	if(handle)
		gnome_vfs_close(handle);
	gnome_vfs_uri_unref(uri);
	return res;
}

static int gnome_vfs_mount_symlink(const char *from, const char *to) {
	int code = 0;
	int res = 0;
	GnomeVFSURI *uri_to = NULL;

	gnome_vfs_mount_message("gnome_vfs_mount_symlink %s -> %s\n", from, to);

	uri_to = gnome_vfs_uri_append_path(base_uri, to);

	code = gnome_vfs_create_symbolic_link(uri_to, from);
	if(code == GNOME_VFS_ERROR_NOT_SUPPORTED)
		res = -EACCES;
	else
	if(code != GNOME_VFS_OK) {
		res = -EACCES;
		gnome_vfs_mount_critical(code, to);
	}

	gnome_vfs_uri_unref(uri_to);
	return res;
}

static int gnome_vfs_mount_rename(const char *from, const char *to) {
	int code = 0;
	int res = 0;
	GnomeVFSURI *uri_from = NULL;
	GnomeVFSURI *uri_to = NULL;

	gnome_vfs_mount_message("gnome_vfs_mount_rename %s -> %s\n", from, to);
	
	uri_from = gnome_vfs_uri_append_path(base_uri, from);
	gnome_vfs_mount_file_info_forget(from);
	uri_to = gnome_vfs_uri_append_path(base_uri, to);

	code = gnome_vfs_move_uri(uri_from, uri_to, TRUE);
	if(code == GNOME_VFS_ERROR_NOT_SUPPORTED)
		res = -EACCES;
	else
	if(code != GNOME_VFS_OK) {
		res = -EACCES;
		gnome_vfs_mount_critical(code, from);
	}

	gnome_vfs_uri_unref(uri_from);
	gnome_vfs_uri_unref(uri_to);
	return res;
}

static int gnome_vfs_mount_chmod(const char *path, mode_t mode) {
	int code = 0;
	int res = 0;
	GnomeVFSURI *uri = NULL;
	GnomeVFSFileInfo *info;

	gnome_vfs_mount_message("gnome_vfs_mount_chmod %s\n", path);
	
	uri = gnome_vfs_uri_append_path(base_uri, path);
	gnome_vfs_mount_file_info_forget(path);
	info = gnome_vfs_file_info_new();

	info->permissions = mode;

	code = gnome_vfs_set_file_info_uri(
				uri, info,
				GNOME_VFS_SET_FILE_INFO_PERMISSIONS);
			
	if(code == GNOME_VFS_ERROR_NOT_SUPPORTED)
		res = -EACCES;
	else
	if(code == GNOME_VFS_ERROR_NOT_PERMITTED)
		res = -EACCES;
	else
	if(code != GNOME_VFS_OK) {
		res = -EACCES;
		gnome_vfs_mount_critical(code, path);
	}
	
	gnome_vfs_uri_unref(uri);
	gnome_vfs_file_info_unref(info);
	return res;
}

static int gnome_vfs_mount_chown(const char *path, uid_t uid, gid_t gid) {
	int code = 0;
	int res = 0;
	GnomeVFSURI *uri = NULL;
	GnomeVFSFileInfo *info = NULL;

	gnome_vfs_mount_message("gnome_vfs_mount_chown %s\n", path);

	uri = gnome_vfs_uri_append_path(base_uri, path);
	gnome_vfs_mount_file_info_forget(path);

	info = gnome_vfs_file_info_new();
	info->uid = uid;
	info->gid = gid;

	code = gnome_vfs_set_file_info_uri(
				uri, info,
				GNOME_VFS_SET_FILE_INFO_OWNER);
			
	if(code == GNOME_VFS_ERROR_NOT_SUPPORTED)
		res = -EACCES;
	else	
	if(code == GNOME_VFS_ERROR_NOT_PERMITTED)
		res = -EACCES;
	else
	if(code != GNOME_VFS_OK) {
		res = -EACCES;
		gnome_vfs_mount_critical(code, path);
	}

	gnome_vfs_uri_unref(uri);
	gnome_vfs_file_info_unref(info);
	return res;
}

static int gnome_vfs_mount_release(const char *path, int flags) {
	gboolean was_found;

	gnome_vfs_mount_message("gnome_vfs_mount_release %s\n", path);

	was_found = g_hash_table_remove(opened_files_hash, path);
	if(was_found == FALSE)
		g_printerr("Trying to release non open path %s\n", path);

	return 0;
}

static struct fuse_operations gnome_vfs_mount_oper = {
    .getattr  = gnome_vfs_mount_getattr,
    .readlink = gnome_vfs_mount_readlink,
    .getdir   = gnome_vfs_mount_getdir,
    .mknod    = gnome_vfs_mount_mknod,
    .mkdir    = gnome_vfs_mount_mkdir,
    .symlink  = gnome_vfs_mount_symlink,
    .unlink   = gnome_vfs_mount_unlink,
    .rmdir    = gnome_vfs_mount_rmdir,
    .rename   = gnome_vfs_mount_rename,
    .chmod    = gnome_vfs_mount_chmod,
    .chown    = gnome_vfs_mount_chown,
    .truncate = gnome_vfs_mount_truncate,
    .utime    = gnome_vfs_mount_utime,
    .open     = gnome_vfs_mount_open,
    .read     = gnome_vfs_mount_read,
    .write    = gnome_vfs_mount_write,
    .release  = gnome_vfs_mount_release
};

static void usage(char *program_name) {
	char *possible_uris[] = {
		"/home/user/libgnomeui-2.8.0.tar.bz2/#bzip2:/#tar:/libgnomeui-2.8.0/",
		"preferences://",
		"sftp://user:password@server/home/user",
		"ftp://ftp.kernel.org/pub"
	};
	int i;
	char usage_s[] = 
		"Usage: %s <gnome-vfs-uri> [options] <mount-directory>\n"
		"\n"
		"Options:\n"
		"  -d, --debug\n"
		"    Enable debug.\n"
		"\n"
		"  -h, -?, --help\n"
		"    Show this help.\n\n"
		"Possible uris are:\n";
	g_print(usage_s, program_name);
	for(i=0; i < sizeof possible_uris / sizeof *possible_uris; i++)
		g_print("\t%s\n", possible_uris[i]);
}

static void gnome_vfs_mount_auth_callback(
		gconstpointer in, size_t in_size, 
		gpointer out, size_t out_size, 
		gpointer user_data) {
	GnomeVFSModuleCallbackFillAuthenticationIn *in_real;
	GnomeVFSModuleCallbackFillAuthenticationOut *out_real;
#ifdef HAVE_GNOME_KEYRING
	GnomeKeyringResult result;
	GList *result_list = NULL;
#endif
				
	in_real = (GnomeVFSModuleCallbackFillAuthenticationIn *)in;
	out_real = (GnomeVFSModuleCallbackFillAuthenticationOut *)out;

#ifdef HAVE_GNOME_KEYRING
	result = gnome_keyring_find_network_password_sync(
				in_real->username,
				in_real->domain,
				in_real->server,
				in_real->object,
				in_real->protocol,
				in_real->authtype,
				in_real->port,
				&result_list);
	
	if (result != GNOME_KEYRING_RESULT_OK || result_list == NULL)
#endif
	{
		char *password = NULL;
		password = getpass("Password: ");
		
		out_real->valid = TRUE;
		out_real->username = g_strdup(in_real->username);
		out_real->password = g_strdup(password);
		memset(password, 0, strlen(password) );
	}
#ifdef HAVE_GNOME_KEYRING	
	else {
		GnomeKeyringNetworkPasswordData *password_data;
		password_data = result_list->data;
		
		out_real->valid = TRUE;
		out_real->username = g_strdup(password_data->user);
		out_real->domain = g_strdup(password_data->domain);
		out_real->password = g_strdup(password_data->password);
		
		gnome_keyring_network_password_list_free(result_list);
	}
#endif
}

static void parse_opt(
		int argc, char *argv[],
		gboolean *show_help, gboolean *enable_debug) {
	struct option long_options[] = {
		{"help", 0, NULL, 'h'},
		{"debug", 0, NULL, 'd'},
		{0, 0, 0, 0}
	};
		
	for(;;) {
		int option_index;
		char opt_c;
		opt_c = getopt_long(argc, argv, "dh?", long_options, &option_index);
		if( opt_c == -1)
			break;
		switch(opt_c) {
			case 'h':
			case '?':
				*show_help = TRUE;
				break;
			case 'd':
				*enable_debug = TRUE;
				break;
		}
	}
}

int main(int argc, char *argv[]) {
	const char *base_uri_s;
	struct stat stbuf;
	char *cmd_argv[3];
	gboolean help = FALSE;
	gboolean debug = FALSE;
	int cmd_argc = 0;
	
	if(argc < 3) {
		usage(argv[0]);
		return 0;
	}

	parse_opt(argc, argv, &help, &debug);	

	if(help) {
		usage(argv[0]);
		return 0;
	}
		
	cmd_argv[0] = argv[0];
	base_uri_s  = argv[optind];
	cmd_argv[1] = argv[optind+1];
	
	if(debug) {
		cmd_argc = 3;
		cmd_argv[2] = "-d";
		__debug = TRUE;
	}
	else {
		cmd_argc = 2;
		__debug = FALSE;
	}
	
	gnome_vfs_mount_init();
	base_uri = gnome_vfs_uri_new(base_uri_s);
	
	gnome_vfs_module_callback_push(
		GNOME_VFS_MODULE_CALLBACK_FILL_AUTHENTICATION,
		gnome_vfs_mount_auth_callback, NULL, NULL);
		
	if( gnome_vfs_uri_exists(base_uri) ) {
		gnome_vfs_module_callback_pop(GNOME_VFS_MODULE_CALLBACK_FILL_AUTHENTICATION);
		gnome_vfs_mount_getattr("/", &stbuf);
		if(stbuf.st_mode & S_IFDIR)
			fuse_main(cmd_argc, cmd_argv, &gnome_vfs_mount_oper);
		else
			g_print("URI is not a directory\n");
	}
	else
		g_print("URI does not exist\n");
	
	gnome_vfs_mount_shutdown();
	return 0;
}
